// --------------------------------------------------------------------------
// functions.c - Sasami functions called from within EDEN
//				 They are called by stub functions in eden/builtin.c, which
//				 are bound to the appropriate EDEN functions in
//				 eden/builtin.h
// --------------------------------------------------------------------------

#include "../config.h"

#ifdef WANT_SASAMI

#include <stdio.h>
#include <stdlib.h>

#include "structures.h"
#include "debug.h"
#include "render.h"

// --------------------------------------------------------------------------
// Called whenever a vertex co-ordinate is updated
// n contains the vertex number, and x,y & z are the new co-ordinates
// --------------------------------------------------------------------------

void sasami_vertex(int n,double x,double y,double z)
{
	vertexinfo *v;

	v=sa_findvertex(n);
	if (v==NULL)
	{
		// Vertex not found so it must be a new one...
#ifdef SASAMI_DEBUG
		printf("Vertex %d created at %g,%g,%g\n",n,x,y,z);
#endif
		v=sa_addvertex(n,x,y,z);
	}
	else
	{
		// Existing vertex - just move it
#ifdef SASAMI_DEBUG
		printf("Vertex %d moved to %g,%g,%g\n",n,x,y,z);
#endif
		v->pos.x=x;
		v->pos.y=y;
		v->pos.z=z;
	}
	sa_r_update(); // Tell the renderer we've changed something
}

// --------------------------------------------------------------------------
// Called whenever the background colour is updated
// --------------------------------------------------------------------------

void sasami_set_bgcolour(double r,double g,double b)
{
	extern colourinfo sa_r_bgcolour;

#ifdef SASAMI_DEBUG
	printf("Background colour set to %g,%g,%g\n",r,g,b);
#endif

	// Update the colour
	sa_r_bgcolour.r=r;
	sa_r_bgcolour.g=g;
	sa_r_bgcolour.b=b;
    sa_r_bgcolour.a=1; // Alpha is always 1 - it makes no sense to have a transparent background!
	// Tell the renderer
	sa_r_update();
}

// --------------------------------------------------------------------------
// Called whenever polygon <p>'s geometry vertex <v> is changed to <id>
// --------------------------------------------------------------------------

void sasami_poly_geom_vertex(int p,int v,int id)
{
	polyinfo	*poly;
	uidlist		*geometry;

	// First find/create polygon <p>

	poly=sa_getpoly(p);

	// Then find/create entry <v> in its UID list

	geometry=sa_getuidlist(&(poly->geometry),v);

	// ...and then set it

	geometry->uid=id;

	// Redraw the display
	sa_r_update();
}

// --------------------------------------------------------------------------
// Called whenever polygon <p>'s texture vertex <v> is changed to <id>
// --------------------------------------------------------------------------

void sasami_poly_tex_vertex(int p,int v,int id)
{
	polyinfo	*poly;
	uidlist		*geometry;

	// First find/create polygon <p>

	poly=sa_getpoly(p);

	// Then find/create entry <v> in its UID list

	geometry=sa_getuidlist(&(poly->texcoords),v);

	// ...and then set it

	geometry->uid=id;

	// Redraw the display
	sa_r_update();
}

// --------------------------------------------------------------------------
// Called whenever polygon <p>'s colour is changed to <r>,<g>,<b>,<a>
// --------------------------------------------------------------------------

void sasami_poly_colour(int p,double r,double g,double b,double a)
{
	polyinfo	*poly;

#ifdef SASAMI_DEBUG
	printf("Setting polygon %d colour to %g,%g,%g,%g\n",p,r,g,b,a);
#endif

	// Find/create polygon <p>

	poly=sa_getpoly(p);

	// Set the colour

	poly->colour.r=r;
	poly->colour.g=g;
	poly->colour.b=b;
	poly->colour.a=a;

	// Redraw the display
	sa_r_update();
}

// --------------------------------------------------------------------------
// Called whenever polygon <p>'s material is changed to <m>
// --------------------------------------------------------------------------

void sasami_poly_material(int p,int m)
{
	polyinfo	*poly;

#ifdef SASAMI_DEBUG
	printf("Setting polygon %d material to %d\n",p,m);
#endif

	// Find/create polygon <p>

	poly=sa_getpoly(p);

	// Set the material

	poly->material=m;

	// Redraw the display
	sa_r_update();
}

// --------------------------------------------------------------------------
// Called whenever object <o>'s polygon <p> is changed to <id>
// --------------------------------------------------------------------------

void sasami_object_poly(int o,int p,int id)
{
	objectinfo	*obj;
	uidlist		*poly;

	// First find/create object <o>

	obj=sa_getobject(o);

	// Then find/create entry <p> in its UID list

	poly=sa_getuidlist(&(obj->polys),p);

	// ...and then set it

	poly->uid=id;

	// Redraw the display
	sa_r_update();
}

// --------------------------------------------------------------------------
// Called whenever object <o>'s position is changed to <x>,<y>,<z>
// --------------------------------------------------------------------------

void sasami_object_pos(int o,double x,double y,double z)
{
	objectinfo	*object;

#ifdef SASAMI_DEBUG
	printf("Setting object %d position to %g,%g,%g\n",o,x,y,z);
#endif

	// Find/create object <o>

	object=sa_getobject(o);

	// Set the position

	object->pos.x=x;
	object->pos.y=y;
	object->pos.z=z;

	// Redraw the display
	sa_r_update();
}

// --------------------------------------------------------------------------
// Called whenever object <o>'s rotation is changed to <x>,<y>,<z>
// --------------------------------------------------------------------------

void sasami_object_rot(int o,double x,double y,double z)
{
	objectinfo	*object;

#ifdef SASAMI_DEBUG
	printf("Setting object %d rotation to %g,%g,%g\n",o,x,y,z);
#endif

	// Find/create object <o>

	object=sa_getobject(o);

	// Set the position

	object->rot.x=x;
	object->rot.y=y;
	object->rot.z=z;

	// Redraw the display
	sa_r_update();
}

// --------------------------------------------------------------------------
// Called whenever object <o>'s scale is changed to <x>,<y>,<z>
// --------------------------------------------------------------------------

void sasami_object_scale(int o,double x,double y,double z)
{
	objectinfo	*object;

#ifdef SASAMI_DEBUG
	printf("Setting object %d scale to %g,%g,%g\n",o,x,y,z);
#endif

	// Find/create object <o>

	object=sa_getobject(o);

	// Set the scale

	object->scale.x=x;
	object->scale.y=y;
	object->scale.z=z;

	// Redraw the display
	sa_r_update();
}

// --------------------------------------------------------------------------
// Called whenever object <o>'s visible status is changed to <v>
// --------------------------------------------------------------------------

void sasami_object_visible(int o,int v)
{
	objectinfo	*object;

#ifdef SASAMI_DEBUG
	printf("Setting object %d visible flag to %d\n",o,v);
#endif

	// Find/create object <o>

	object=sa_getobject(o);

	// Set the flag

	object->visible=v;

	// Redraw the display
	sa_r_update();
}

// --------------------------------------------------------------------------
// Called whenever the viewport size is changed
// --------------------------------------------------------------------------

void sasami_viewport(int x,int y)
{
#ifdef SASAMI_DEBUG
	printf("Setting viewport size to %dx%d\n",x,y);
#endif

	sa_r_resizeviewport(x,y);
}

// --------------------------------------------------------------------------
// Called whenever the "show axes" state is changed
// --------------------------------------------------------------------------

void sasami_setshowaxes(int n)
{
#ifdef SASAMI_DEBUG
	printf("Setting axes display to %d\n",n);
#endif

	sa_r_setshowaxes(n);
}

// --------------------------------------------------------------------------
// Called whenever material <m>'s ambient is changed to <r>,<g>,<b>,<a>
// --------------------------------------------------------------------------

void sasami_material_ambient(int m,double r,double g,double b,double a)
{
	materialinfo	*material;

#ifdef SASAMI_DEBUG
	printf("Setting material %d ambient to %g,%g,%g,%g\n",m,r,g,b,a);
#endif

	// Find/create material <m>

	material=sa_getmaterial(m);

	// Set the colour

	material->ambient.r=r;
	material->ambient.g=g;
	material->ambient.b=b;
	material->ambient.a=a;

	// Redraw the display
	sa_r_update();
}

// --------------------------------------------------------------------------
// Called whenever material <m>'s diffuse is changed to <r>,<g>,<b>,<a>
// --------------------------------------------------------------------------

void sasami_material_diffuse(int m,double r,double g,double b,double a)
{
	materialinfo	*material;

#ifdef SASAMI_DEBUG
	printf("Setting material %d diffuse to %g,%g,%g,%g\n",m,r,g,b,a);
#endif

	// Find/create material <m>

	material=sa_getmaterial(m);

	// Set the colour

	material->diffuse.r=r;
	material->diffuse.g=g;
	material->diffuse.b=b;
	material->diffuse.a=a;

	// Redraw the display
	sa_r_update();
}

// --------------------------------------------------------------------------
// Called whenever material <m>'s specular is changed to <r>,<g>,<b>,<a>
// --------------------------------------------------------------------------

void sasami_material_specular(int m,double r,double g,double b,double a)
{
	materialinfo	*material;

#ifdef SASAMI_DEBUG
	printf("Setting material %d specular to %g,%g,%g,%g\n",m,r,g,b,a);
#endif

	// Find/create material <m>

	material=sa_getmaterial(m);

	// Set the colour

	material->specular.r=r;
	material->specular.g=g;
	material->specular.b=b;
	material->specular.a=a;

	// Redraw the display
	sa_r_update();
}

// --------------------------------------------------------------------------
// Called whenever material <m>'s texture is changed to <t>
// --------------------------------------------------------------------------

void sasami_material_texture(int m,char *t)
{
	materialinfo	*material;

#ifdef SASAMI_DEBUG
	printf("Setting material %d texture to \"%s\"\n",m,t);
#endif

	// Find/create material <m>

	material=sa_getmaterial(m);

	// Set the texture

	material->texturefile=realloc(material->texturefile,strlen(t)+1);
	strcpy(material->texturefile,t);

	// Set the texture binding to 0 and tell the renderer to load it
	// (we can't load texture now as OpenGL may not be initialised!)

	material->texture=0;
	material->textureneedsloading=1;

	// Redraw the display

	sa_r_update();
}

// --------------------------------------------------------------------------
// Called whenever light <o>'s position is changed to <x>,<y>,<z>
// --------------------------------------------------------------------------

void sasami_light_pos(int o,double x,double y,double z)
{
	lightinfo	*light;

#ifdef SASAMI_DEBUG
	printf("Setting light %d position to %g,%g,%g\n",o,x,y,z);
#endif

	// Find/create light <o>

	light=sa_getlight(o);

	// Set the position

	light->pos.x=x;
	light->pos.y=y;
	light->pos.z=z;

	// Redraw the display
	sa_r_update();
}

// --------------------------------------------------------------------------
// Called whenever light <m>'s ambient is changed to <r>,<g>,<b>,<a>
// --------------------------------------------------------------------------

void sasami_light_ambient(int m,double r,double g,double b,double a)
{
	lightinfo	*light;

#ifdef SASAMI_DEBUG
	printf("Setting light %d ambient to %g,%g,%g,%g\n",m,r,g,b,a);
#endif

	// Find/create light <m>

	light=sa_getlight(m);

	// Set the colour

	light->ambient.r=r;
	light->ambient.g=g;
	light->ambient.b=b;
	light->ambient.a=a;

	// Redraw the display
	sa_r_update();
}

// --------------------------------------------------------------------------
// Called whenever light <m>'s diffuse is changed to <r>,<g>,<b>,<a>
// --------------------------------------------------------------------------

void sasami_light_diffuse(int m,double r,double g,double b,double a)
{
	lightinfo	*light;

#ifdef SASAMI_DEBUG
	printf("Setting light %d diffuse to %g,%g,%g,%g\n",m,r,g,b,a);
#endif

	// Find/create light <m>

	light=sa_getlight(m);

	// Set the colour

	light->diffuse.r=r;
	light->diffuse.g=g;
	light->diffuse.b=b;
	light->diffuse.a=a;

	// Redraw the display
	sa_r_update();
}

// --------------------------------------------------------------------------
// Called whenever light <m>'s specular is changed to <r>,<g>,<b>,<a>
// --------------------------------------------------------------------------

void sasami_light_specular(int m,double r,double g,double b,double a)
{
	lightinfo	*light;

#ifdef SASAMI_DEBUG
	printf("Setting light %d specular to %g,%g,%g,%g\n",m,r,g,b,a);
#endif

	// Find/create light <m>

	light=sa_getlight(m);

	// Set the colour

	light->specular.r=r;
	light->specular.g=g;
	light->specular.b=b;
	light->specular.a=a;

	// Redraw the display
	sa_r_update();
}

// --------------------------------------------------------------------------
// Called whenever light <m>'s enabled status is changed to <e>
// --------------------------------------------------------------------------

void sasami_light_enabled(int m,int e)
{
	lightinfo	*light;

#ifdef SASAMI_DEBUG
	printf("Setting light %d enabled flag to %d\n",m,e);
#endif

	// Find/create light <m>

	light=sa_getlight(m);

	// Set the flag

	light->enabled=e;

	// Redraw the display
	sa_r_update();
}

// --------------------------------------------------------------------------
// Called whenever light <m>'s attenuation is changed to <a>
// --------------------------------------------------------------------------

void sasami_light_attenuation(int m,double a)
{
	lightinfo	*light;

#ifdef SASAMI_DEBUG
	printf("Setting light %d attenuation to %g\n",m,a);
#endif

	// Find/create light <m>

	light=sa_getlight(m);

	// Set the attenuation

	light->attenuation=a;

	// Redraw the display
	sa_r_update();
}

// --------------------------------------------------------------------------
// Called whenever light <m>'s directional status is changed to <d>
// --------------------------------------------------------------------------

void sasami_light_directional(int m,int d)
{
	lightinfo	*light;

#ifdef SASAMI_DEBUG
	printf("Setting light %d directional flag to %d\n",m,d);
#endif

	// Find/create light <m>

	light=sa_getlight(m);

	// Set the flag

	light->directional=d;

	// Redraw the display
	sa_r_update();
}

#endif /* WANT_SASAMI */
